# JSFragments

JavaScript library for rendering annotations.

## How To

[[_TOC_]]

### HTML file structure

```html
<html lang="en" dir="ltr">
  <head>
    <meta charset="utf-8">
    <title>Test fragments</title>
    <script src="fragments.js" charset="utf-8"></script>
    <script type="text/javascript">
$(document).ready(function () {
  // fragments definition goes here
}
    </script>
  </head>
  <body>
    <div id="frags"><!-- fragments rendering goes here -->
    </div>
  </body>
</html>
```

**Attention**: the inner content of `div#frags` will be replaced with the fragment rendering. This this element should be empty.

### Fragment Container

The fragment container is a container object of class *FragmentContainer* where all fragments will be stored.

```js
document.addEventListener('DOMContentLoaded', function () {
  const name = 'example';
  const text = 'Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed non risus.';
  const fcont = new FragmentContainer(name, text);
});
```

- *name* is the name of the container.
- *text* is the text content of the container. Fragment offsets stored in the container will be relative to this text.

### Layout fragments

Layout fragments are fragments that represent standard HTML tags like `<em>`, `<i>`, or `<p>`. You can add a layout fragment with the `addLayout()` method.

```js
  fcont.addLayout('p', 0, 56);
  fcont.addLayout('p', 57, 71);
  fcont.addLayout('i', 0, 11);
  fcont.addLayout('b', 61, 64);
  fcont.addLayout('a', 28, 39, {attributes: {href: 'https://en.wiktionary.org/wiki/consectetur'}});
```

This example creates two paragraphs, an italic text, a bold text and a link anchor.

### Fragment builder

A fragment builder renders a fragment container text content and all the fragments stored.

```js
  const styler = new FragmentBoxColorMentionStyler();
  const builder = fcont.addBuilder('#frags', FragmentBuilder, styler);
  fcont.build();
```

The `addBuilder()` method creates a builder for the fragment container. This method accepts three arguments:
- *targetElement*: the element where to render the builder. This argument may be a DOM object or a selector string.
- *builderClass*: the type of rendering. Here we will go for the *FragmentBuilder*. Later we will see other classes of builders.
- *styler*: a styling object that controls the looks of the rendered text and fragments. We will explore this object later on.

The `build()` method tells the fragment container to render all its builders. If you open the HTML file in a browser, you will see something like this:

<details><summary><i>Layout fragments rendering</i></summary>
![Layout fragments rendering](scrot-layout.png)
</details>

All layout fragments are rendered as intended.

### Mention fragment

Mention fragments represent mentions or named entities of interest. Mentions are rendered as `<span>` with a different look than layout fragments.

Add the following *before* creating the styler:

```js
  fcont.addMention('Person', 0, 11);
  fcont.addMention('Disease', 6, 17);
  fcont.addMention('Joke', 57, 71);
  fcont.addMention('Something', 28, 55);
  fcont.addMention('Joke', 40, 50);
```

The `addMention()` method creates a new mention in the container. The first argument is the mention *type*, you decide which types suits your needs.

The rendering looks like this:

<details><summary><i>Mention rendering</i></summary>
![Mention rendering](scrot-mention-grey.png)
</details>

Mentions are rendered in grey boxes. Notice how it meshes well with layout fragments, and how it renders embedded and overlapping mentions.

### Colors

In the previous example, all mentions are rendered in gray. That is sad and does not allow the reader to distinguish between mention types.

Let's tell the styler to add some colors:

```js
  const colors = Colors.spread();
  const styler = new FragmentBoxColorMentionStyler({colors: colors});
```

<details><summary><i>Color rendering</i></summary>
![Color rendering](scrot-mention-color.png)
</details>

Here we told the styler to chose one color for each mention type in a way that hues are spread around the color wheel. You can also customize the saturation, lightness and the hue of the first color:

```js
  const colors = Colors.spread(-77, 100, 60);
  const styler = new FragmentBoxColorMentionStyler({colors: colors});
```

Alternatively, you might ask for random hues:

```js
  const colors = Colors.random();
  const styler = new FragmentBoxColorMentionStyler({colors: colors});
```

### Underline style

Instead of rendering mentions as boxes, you can render them as colored underlines by specifying another styler:

```js
  const styler = new FragmentUnderlineColorMentionStyler({colors: colors});
```

<details><summary><i>Underline rendering</i></summary>
![Underline rendering](scrot-mention-underline.png)
</details>

### Highlight a particular mention

The fragment container `highlight()` method allows to draw the attention to a particular mention.

Add this after rendering:

```js
  fcont.highlight(fcont.mentions[2]);
```

<details><summary><i>Mention highlighting</i></summary>
![Mention highlighting](scrot-mention-highlight.png)
</details>

If you try to highlight several mentions like this:

```js
  fcont.highlight(fcont.mentions[2]);
  fcont.highlight(fcont.mentions[4]);
```

You will realize that only the second mention will be highlighted. That's because by default `highlight()` removes previously highlighted mentions. If you want to highlight a second mention, try this:

```js
  fcont.highlight(fcont.mentions[2]);
  fcont.highlight(fcont.mentions[4], false);
```

The second argument of `highlight()` tells the container not to remove previous highlights.

### Highlight clicked mention

The `Mention.from()` static method retrieves a mention object from an element or an event. This method allows you to get the mention that was clicked on.

```js
  builder.buildRoot.onclick = function(event) { fcont.highlight(Mention.from(event)); };
```

This snippet will highlight the clicked mention.
